/**
 * Copyright (c) 2012 the Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/


#include "./intersection_detection.h"

#include <assert.h>

#include "./line.h"
#include "./vec.h"

#define MIN2(a, b) ((a < b) ? (a): (b))
#define MIN4(a, b, c, d) MIN2(MIN2(a, b), MIN2(c, d))
#define MAX2(a, b) ((a > b) ? (a): (b))
#define MAX4(a, b, c, d) MAX2(MAX2(a, b), MAX2(c, d))

#define LESS_THAN_MIN4(x, a, b, c, d) (x < a && x < b && x < c && x < d)
#define GREATER_THAN_MAX4(x, a, b, c, d) (x > a && x > b && x > c && x > d)
#define X_SEPERABLE_LEFT(p1, p2, p3, p4) (p1.x < p3.x && p1.x < p4.x \
                                       && p2.x < p3.x && p2.x < p4.x)
#define Y_SEPERABLE_DOWN(p1, p2, p3, p4) (p1.y < p3.y && p1.y < p4.y \
                                       && p2.y < p3.y && p2.y < p4.y)

#define SAME_SIGN(d1, d2) (d1 > 0 && d2 > 0 || d1 < 0 && d2 < 0)

void update_bounding_box(Line *l1, double time) {
  l1->bounding_box.target_p1 = Vec_add(l1->p1, Vec_multiply(l1->velocity, time));
  l1->bounding_box.target_p2 = Vec_add(l1->p2, Vec_multiply(l1->velocity, time));

  l1->bounding_box.xmax = MAX4(l1->p1.x, l1->p2.x,
                  l1->bounding_box.target_p1.x, l1->bounding_box.target_p2.x);
  l1->bounding_box.xmin = MIN4(l1->p1.x, l1->p2.x,
                  l1->bounding_box.target_p1.x, l1->bounding_box.target_p2.x);
  l1->bounding_box.ymax = MAX4(l1->p1.y, l1->p2.y,
                  l1->bounding_box.target_p1.y, l1->bounding_box.target_p2.y);
  l1->bounding_box.ymin = MIN4(l1->p1.y, l1->p2.y,
                  l1->bounding_box.target_p1.y, l1->bounding_box.target_p2.y);
}


// Detect if lines l1 and l2 will intersect between now and the next time step.
IntersectionType intersect(Line *l1, Line *l2, double time) {
  assert(compareLines(l1, l2) < 0);

  // eliminate separable case
  if (l1->bounding_box.xmax < l2->bounding_box.xmin
   || l1->bounding_box.xmin > l2->bounding_box.xmax
   || l1->bounding_box.ymax < l2->bounding_box.ymin
   || l1->bounding_box.ymin > l2->bounding_box.ymax) {
    return NO_INTERSECTION;
  }

  // Relative position of target_p1 and target_p2 of l2
  // when we fix the velocity of l1 to be 0
  Vec rel_target_p1 = Vec_add(l2->bounding_box.target_p1, 
                      Vec_subtract(l1->p1, l1->bounding_box.target_p1));
  Vec rel_target_p2 = Vec_add(l2->bounding_box.target_p2,
                      Vec_subtract(l1->p2, l1->bounding_box.target_p2));

  int num_line_intersections = 0;
  bool top_intersected = false;
  bool bottom_intersected = false;

  if (intersectLines(l1->p1, l1->p2, l2->p1, l2->p2)) {
    return ALREADY_INTERSECTED;
  }
  if (intersectLines(l1->p1, l1->p2, rel_target_p1, rel_target_p2)) {
    num_line_intersections++;
  }
  if (intersectLines(l1->p1, l1->p2, rel_target_p1, l2->p1)) {
    num_line_intersections++;
    top_intersected = true;
  }
  if (intersectLines(l1->p1, l1->p2, rel_target_p2, l2->p2)) {
    num_line_intersections++;
    bottom_intersected = true;
  }

  if (num_line_intersections == 2) {
    return L2_WITH_L1;
  }

  if(pointInParallelogram(l1->p1, l2->p1, l2->p2, rel_target_p1, rel_target_p2)
   && pointInParallelogram(l1->p2, l2->p1, l2->p2, rel_target_p1, rel_target_p2)) {
    return L1_WITH_L2;
  }

  if (num_line_intersections == 0) {
    return NO_INTERSECTION;
  }

  // We don;t need to calculate angle in advance
  // It happens at most once
//  Vec v1 = Vec_makeFromLine(*l1);
//  Vec v2 = Vec_makeFromLine(*l2);
//  double angle = Vec_angle(v1, v2);

  if (top_intersected) {
    Vec v1 = Vec_makeFromLine(*l1);
    Vec v2 = Vec_makeFromLine(*l2);
    if (Vec_angle(v1, v2) < 0) {
      return L2_WITH_L1;
    } else {
      return L1_WITH_L2;
    }
  }

  if (bottom_intersected) {
    Vec v1 = Vec_makeFromLine(*l1);
    Vec v2 = Vec_makeFromLine(*l2);
    if (Vec_angle(v1, v2) > 0) {
      return L2_WITH_L1;
    } else {
      return L1_WITH_L2;
    }
  }

  return L1_WITH_L2;
}

// Check if a point is in the parallelogram.
bool pointInParallelogram(Vec point, Vec p1, Vec p2, Vec p3, Vec p4) {
  // Eliminating simple cases to reduce calculations
/*
  if (point.x < p1.x && point.x < p2.x && point.x < p3.x && point.x < p4.x) {
    return false;
  }
  if (point.x > p1.x && point.x > p2.x && point.x > p3.x && point.x > p4.x) {
    return false;
  }
  if (point.y < p1.y && point.y < p2.y && point.y < p3.y && point.y < p4.y) {
    return false;
  }
  if (point.y > p1.y && point.y > p2.y && point.y > p3.y && point.y > p4.y) {
    return false;
  }
*/
  if (LESS_THAN_MIN4(point.x, p1.x, p2.x, p3.x, p4.x)){
    return false;
  } 
  if (GREATER_THAN_MAX4(point.x, p1.x, p2.x, p3.x, p4.x)){
    return false;
  } 
  if (LESS_THAN_MIN4(point.y, p1.y, p2.y, p3.y, p4.y)){
    return false;
  } 
  if (GREATER_THAN_MAX4(point.y, p1.y, p2.y, p3.y, p4.y)){
    return false;
  } 

  double d1 = direction(p1, p2, point);
  double d2 = direction(p3, p4, point);
  double d3 = direction(p1, p3, point);
  double d4 = direction(p2, p4, point);

  if (((d1 > 0 && d2 < 0) || (d1 < 0 && d2 > 0))
      && ((d3 > 0 && d4 < 0) || (d3 < 0 && d4 > 0))) {
    return true;
  }
  return false;
}

// Check if two lines intersect.
bool intersectLines(Vec line1start, Vec line1end, Vec line2start, Vec line2end) {
  // Eliminating simple cases to reduce calculation

  if (X_SEPERABLE_LEFT(line1start, line1end, line2start, line2end)) {
    return false;
  }
  if (X_SEPERABLE_LEFT(line2start, line2end, line1start, line1end)) {
    return false;
  }
  if (Y_SEPERABLE_DOWN(line1start, line1end, line2start, line2end)) {
    return false;
  }
  if (Y_SEPERABLE_DOWN(line2start, line2end, line1start, line1end)) {
    return false;
  }
/*
  if (p1.x < p3.x && p1.x < p4.x && p2.x < p3.x && p2.x < p4.x) {
    return false;
  }
  if (p1.x > p3.x && p1.x > p4.x && p2.x > p3.x && p2.x > p4.x) {
    return false;
  } 
  if (p1.y < p3.y && p1.y < p4.y && p2.y < p3.y && p2.y < p4.y) {
    return false;
  } 
  if (p1.y > p3.y && p1.y > p4.y && p2.y > p3.y && p2.y > p4.y) {
    return false;
  }
*/
  // Relative orientation
  double d1 = direction(line2start, line2end, line1start);
  double d2 = direction(line2start, line2end, line1end);
  double d3 = direction(line1start, line1end, line2start);
  double d4 = direction(line1start, line1end, line2end);

  // If two points on the line is on the same side, the line segments
  // never intersect.
  if (SAME_SIGN(d1, d2) || SAME_SIGN(d3, d4)) {
    return false;
  }
  // If (line1start, line1end) and (line2start, line2end) straddle each other, the line segments must
  // intersect.
  if (((d1 > 0 && d2 < 0) || (d1 < 0 && d2 > 0))
      && ((d3 > 0 && d4 < 0) || (d3 < 0 && d4 > 0))) {
    return true;
  }
  if (d1 == 0 || d2 == 0) {
    return (d1 == 0 && onSegment(line2start, line2end, line1start)) || (d2 == 0 && onSegment(line2start, line2end, line1end));
  } else {
    return (d3 == 0 && onSegment(line1start, line1end, line2start)) || (d4 == 0 && onSegment(line1start, line1end, line2end));
   }


/*
  // Remaining case: there exists three points that are colinear  
  if (d1 == 0 || d2 == 0) {
    return (d1 == 0 && onSegment(line2start, line2end, line1start)) || (d2 == 0 && onSegment(line2start, line2end, line1end));
  } else {
    return (d3 == 0 && onSegment(line1start, line1end, line2end)) || (d4 == 0 && onSegment(line1start, line1end, line2end));
  }
*/
}

// Obtain the intersection point for two intersecting line segments.
Vec getIntersectionPoint(Vec p1, Vec p2, Vec p3, Vec p4) {
  double u;

  u = ((p4.x - p3.x) * (p1.y - p3.y) - (p4.y - p3.y) * (p1.x - p3.x))
      / ((p4.y - p3.y) * (p2.x - p1.x) - (p4.x - p3.x) * (p2.y - p1.y));

  return Vec_add(p1, Vec_multiply(Vec_subtract(p2, p1), u));
}

// Check the direction of two lines (pi, pj) and (pi, pk).
double direction(Vec pi, Vec pj, Vec pk) {
  return crossProduct(pk.x - pi.x, pk.y - pi.y, pj.x - pi.x, pj.y - pi.y);
}

// Check if a point pk is in the line segment (pi, pj).
// pi, pj, and pk must be collinear.
bool onSegment(Vec pi, Vec pj, Vec pk) {
  assert(direction(pi, pj, pk) == 0);
  // we simplify the arithmetic expression to end up checking this value. This
  // is valid because when the 3 points are colinear, if y coordinates are between
  // those 2 lines, then the x coordinates are between those 2 lines
  if ((pi.x <= pk.x && pk.x <= pj.x) || (pj.x <= pk.x && pk.x <= pi.x)) {
    return true;
  }
  return false;
}

// Calculate the cross product.
double crossProduct(double x1, double y1, double x2, double y2) {
  return x1 * y2 - x2 * y1;
}

