#include "./line.h"
#include "./quadtree.h"
#include "./collision_world.h"
#include "./intersection_event_list.h"
#include "./line_linked_list.h"

#include "./fasttime.h"
#include <stdio.h>

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <cilk/cilk.h>
#include <cilk/reducer.h>
#include <cilk/reducer_opadd.h>

#define THRESHOLD 150
#define MAXDEPTH 4
#define LOOPTHRESHOLD 20

#define ALLOCATE(node) node->q1 = malloc(sizeof(QuadTree)); \
  node->q2 = malloc(sizeof(QuadTree)); \
  node->q3 = malloc(sizeof(QuadTree)); \
  node->q4 = malloc(sizeof(QuadTree)); \


#define INITIALIZE(node) node->initialized = true; \
  node->q1->initialized = false; \
  node->q2->initialized = false; \
  node->q3->initialized = false; \
  node->q4->initialized = false; \

/*
// clears the Line points of root
void unlink_list(QuadTree * root){
  root->first = NULL;
  root->last = NULL;
  root->line_length = 0;
}

// adds a line to the root's Line pointers
void add_line(QuadTree * root, Line * line){
  line->next = NULL;
  root->line_length++;
  // Check whether ths line is the first line attached.
  if (root->first == NULL){
    root->first = line;
    root->last = line;
    return;
  }
  root->last->next = line;
  root->last = line;
}
*/

typedef CILK_C_DECLARE_REDUCER(IntersectionEventList) IntersectionEventListReducer;
CILK_C_REDUCER_OPADD(numLineLineCollisionReducer, int, 0);

void intersection_event_list_init(IntersectionEventList* intersectionEventList) {
  intersectionEventList->head = NULL;
  intersectionEventList->tail = NULL;
}

void num_line_line_collision_init(int* num) {
  *num = 0;
}

// merges two IntersectionEventList's
void intersection_event_list_reduce(void* key, void* left_void, void* right_void) {
  IntersectionEventList* left = (IntersectionEventList*) left_void;
  IntersectionEventList* right = (IntersectionEventList*) right_void;

  if (right->head == NULL) {
    assert(right->tail == NULL);
    return;
  }
  if (left->head == NULL) {
    assert(left->tail == NULL);
    left->head = right->head;
    left->tail = right->tail;
    right->head = NULL;
    right->tail = NULL;
    return;
  }
  left->tail->next = right->head;
  left->tail = right->tail;
  right->head = NULL;
  right->tail = NULL;
}

void intersection_event_list_identity(void* key, void* value) {
  intersection_event_list_init((IntersectionEventList*) value);
}

void intersection_event_list_destroy(void* key, void* value) {
  IntersectionEventList_deleteNodes((IntersectionEventList*) value);
}

IntersectionEventListReducer intersectionEventListReducer = CILK_C_INIT_REDUCER(IntersectionEventList, intersection_event_list_reduce, intersection_event_list_identity, intersection_event_list_destroy, (IntersectionEventList) { .head = NULL, .tail = NULL });


// Makes a QuadTree root, calling make_quadtree_node for its 4 subchildren.
// Counts intersections with lines contained at the root and other lines at the root and lines
// contains in subtrees via call to count_collision_now
void make_quadtree_root(QuadTree* root, CollisionWorld* collisionWorld) {  
  root->depth = 0;
  ALLOCATE(root);
  if (!(root->q1 && root->q2 && root->q3 && root->q4)) {
    return;
  }
  INITIALIZE(root); 
  root->q1->depth = 1;
  root->q2->depth = 1;
  root->q3->depth = 1;
  root->q4->depth = 1;
/*
  root->q1->lineArray = NULL;
  root->q2->lineArray = NULL;
  root->q3->lineArray = NULL;
  root->q4->lineArray = NULL;
*/

  // Initialize first and last in quadtress to be NULL
  unlink_list(&root->lineLinkedList);
  unlink_list(&root->q1->lineLinkedList);
  unlink_list(&root->q2->lineLinkedList);
  unlink_list(&root->q3->lineLinkedList);
  unlink_list(&root->q4->lineLinkedList);

  // Assign lines to their corresponding quadrants if they fit in 1,
  // or keep lines at this QuadTree if they don't fit
  for (int i = 0; i < collisionWorld->numOfLines; i++) {
    Line * line = collisionWorld->lines[i];
    which_quadrant_line(line, BOX_XMIN, BOX_XMAX, BOX_YMIN, BOX_YMAX, collisionWorld->timeStep); 
    add_line_to_quadrant(line, root);
  }

  // Count the possible number of collisions with the lines in the root
//  count_collision_now(root, collisionWorld, intersection_event_list_reducer);
 
  // Recursively build quadtrees
  cilk_spawn make_quadtree_node(root->q1, collisionWorld, (BOX_XMIN + BOX_XMAX) / 2, BOX_XMAX, (BOX_YMAX + BOX_YMIN) / 2, BOX_YMAX); 
  cilk_spawn make_quadtree_node(root->q2, collisionWorld, BOX_XMIN, (BOX_XMIN + BOX_XMAX) / 2, (BOX_YMAX + BOX_YMIN) / 2, BOX_YMAX); 
  cilk_spawn make_quadtree_node(root->q3, collisionWorld, BOX_XMIN, (BOX_XMIN + BOX_XMAX) / 2, BOX_YMIN, (BOX_YMAX + BOX_YMIN) / 2); 
  make_quadtree_node(root->q4, collisionWorld, (BOX_XMIN + BOX_XMAX) / 2, BOX_XMAX, BOX_YMIN, (BOX_YMAX + BOX_YMIN) / 2); 

//  root->lineArray = malloc(root->lineLinkedList.length * sizeof(Line *));
//  Line* l1 = root->lineLinkedList.first;
// for (size_t i = 0; i < root->lineLinkedList.length; i++) {
//    root->lineArray[i] = l1;
//    l1 = l1->next;
//  }
//  cilk_sync;
}

// Recursively builds a QuadTree node, and counts the number of collisions between lines contained at this node and other lines at this node
// and lines contained in subtrees
void make_quadtree_node(QuadTree* node, CollisionWorld* collisionWorld, const vec_dimension left, const vec_dimension right, const vec_dimension bottom, const vec_dimension top){  
  node->q1 = NULL;
  node->q2 = NULL;
  node->q3 = NULL;
  node->q4 = NULL;

  // If the number of lines is less then THRESHOLD
  // or we reach the maximum depth, stop build the tree and 
  // count the number of collisions
  if (node->lineLinkedList.length <= THRESHOLD || node->depth >= MAXDEPTH) {
//    count_collision_now(node, collisionWorld, intersection_event_list_reducer);
    return;
  }

  // Initialization
  ALLOCATE(node);
  if (!(node->q1 && node->q2 && node->q3 && node->q4)) {
    return;
  }
  INITIALIZE(node);
  node->q1->depth = node->depth + 1;
  node->q2->depth = node->depth + 1;
  node->q3->depth = node->depth + 1;
  node->q4->depth = node->depth + 1;
/*
  node->q1->lineArray = NULL;
  node->q2->lineArray = NULL;
  node->q3->lineArray = NULL;
  node->q4->lineArray = NULL;
*/
  // Get the line attached first to the node before unlink it
  Line * line = node->lineLinkedList.first;
  int line_length = node->lineLinkedList.length;
  unlink_list(&node->lineLinkedList);
  unlink_list(&node->q1->lineLinkedList);
  unlink_list(&node->q2->lineLinkedList);
  unlink_list(&node->q3->lineLinkedList);
  unlink_list(&node->q4->lineLinkedList);


  // Assign lines to their corresponding quadrants if they fit in 1,
  // or keep lines at this QuadTree if they don't fit
  for(int i = 0; i < line_length; i++){
    Line * next = line->next;
    which_quadrant_line(line, left, right, bottom, top, collisionWorld->timeStep);

    add_line_to_quadrant(line, node);
    line = next;
  }

  // Count the possible number of collisions with the lines in the root
//  count_collision_now(node, collisionWorld, intersection_event_list_reducer);

  
  // Recursively build quadtrees
  if (node->depth < MAXDEPTH - 2) {
    cilk_spawn  make_quadtree_node(node->q1, collisionWorld, (left + right) / 2, right, (bottom + top) / 2, top); 
    cilk_spawn make_quadtree_node(node->q2, collisionWorld, left, (left + right) / 2, (bottom + top) / 2, top); 
    cilk_spawn make_quadtree_node(node->q3, collisionWorld, left, (left + right) / 2, bottom, (bottom + top) / 2); 
    make_quadtree_node(node->q4, collisionWorld, (left + right) / 2, right, bottom, (bottom + top) / 2); 
  } else {
    make_quadtree_node(node->q1, collisionWorld, (left + right) / 2, right, (bottom + top) / 2, top); 
    make_quadtree_node(node->q2, collisionWorld, left, (left + right) / 2, (bottom + top) / 2, top); 
    make_quadtree_node(node->q3, collisionWorld, left, (left + right) / 2, bottom, (bottom + top) / 2); 
    make_quadtree_node(node->q4, collisionWorld, (left + right) / 2, right, bottom, (bottom + top) / 2);  
  }

//  node->lineArray = malloc(node->lineLinkedList.length * sizeof(Line *));
//  Line* l1 = node->lineLinkedList.first;
//  for (size_t i = 0; i < node->lineLinkedList.length; i++) {
//    node->lineArray[i] = l1;
//    l1 = l1->next;
//  }
//  cilk_sync;
}

void cleanup_quadtree(QuadTree * node){
  unlink_list(&node->lineLinkedList);

  if (node->q1 != NULL) {
    cleanup_quadtree(node->q1);
  }
  if (node->q2 != NULL) {
    cleanup_quadtree(node->q2);
  }
  if (node->q3 != NULL) {
    cleanup_quadtree(node->q3);
  }
  if (node->q4 != NULL) {
    cleanup_quadtree(node->q4);
  }
}

// Updates a QuadTree root (with already allocated memory), calling update_quadtree_node for its 4 subchildren
// It clears the lines contained at this root and recursively rebuilds the lines.
// Counts intersections with lines contained at the root and other lines at the root and lines
// contains in subtrees via call to count_collision_now
// Different from make_quadtree_root because this function does not reallocate memory for the root
void update_quadtree_root(QuadTree* root, CollisionWorld* collisionWorld) {
  cleanup_quadtree(root);

  for (int i = 0; i < collisionWorld->numOfLines; i++) {
    Line * line = collisionWorld->lines[i];
    which_quadrant_line(line, BOX_XMIN, BOX_XMAX, BOX_YMIN, BOX_YMAX, collisionWorld->timeStep);
 
    add_line_to_quadrant(line, root);
  }
  
  // Count the number of collisions with the lines in the root
//  count_collision_now(root, collisionWorld, intersection_event_list_reducer);

  // Recursively update quadtrees
  cilk_spawn update_quadtree_node(root->q1, collisionWorld, (BOX_XMIN + BOX_XMAX) / 2, BOX_XMAX, (BOX_YMAX + BOX_YMIN) / 2, BOX_YMAX); 
  cilk_spawn update_quadtree_node(root->q2, collisionWorld, BOX_XMIN, (BOX_XMIN + BOX_XMAX) / 2, (BOX_YMAX + BOX_YMIN) / 2, BOX_YMAX); 
  cilk_spawn update_quadtree_node(root->q3, collisionWorld, BOX_XMIN, (BOX_XMIN + BOX_XMAX) / 2, BOX_YMIN, (BOX_YMAX + BOX_YMIN) / 2); 
  update_quadtree_node(root->q4, collisionWorld, (BOX_XMIN + BOX_XMAX) / 2, BOX_XMAX, BOX_YMIN, (BOX_YMAX + BOX_YMIN) / 2); 

//  if (root->lineArray != NULL) {
//    free(root->lineArray);
//  }
//  root->lineArray = malloc(root->lineLinkedList.length * sizeof(Line *));
//  Line* l1 = root->lineLinkedList.first;
//  for (size_t i = 0; i < root->lineLinkedList.length; i++) {
//    root->lineArray[i] = l1;
//    l1 = l1->next;
//  }
 
  cilk_sync;
}



// Recursively updates a QuadTree node, and counts the number of collisions between lines contained at this node and other lines at this node
// and lines contained in subtrees.
// Different from make_quadtree_node because memory for the node is already allocated beforehand.
void update_quadtree_node(QuadTree* node, CollisionWorld* collisionWorld, const vec_dimension left, const vec_dimension right, const vec_dimension bottom, const vec_dimension top){
  // If the node has not been initialized, call make_quadtree_node
  // to build tress from it.
  if (!node->initialized){
    make_quadtree_node(node, collisionWorld, left, right, bottom, top);
    return;
  }
  
  // If the number of lines is less then THRESHOLD
  // or we reach the maximum depth, stop build the tree and 
  // count the number of collisions
  if (node->lineLinkedList.length <= THRESHOLD || node->depth >= MAXDEPTH) {
//    count_collision_now(node, collisionWorld, intersection_event_list_reducer);
    return;
  }

  // Get the first line attached before unlink it.
  Line * line = node->lineLinkedList.first;
  int line_length = node->lineLinkedList.length;
  unlink_list(&node->lineLinkedList);

  for (int i = 0; i < line_length; i++) {
    which_quadrant_line(line, left, right, bottom, top, collisionWorld->timeStep);
    Line * next = line->next;

    add_line_to_quadrant(line, node);
    line = next;
  }

  // Count the number of collisions with the lines in this node
//  count_collision_now(node, collisionWorld, intersection_event_list_reducer);
 
  // Recursively update quadtrees
  if (node->depth < MAXDEPTH - 2) {
    cilk_spawn update_quadtree_node(node->q1, collisionWorld, (left + right) / 2, right, (bottom + top) / 2, top); 
    cilk_spawn update_quadtree_node(node->q2, collisionWorld, left, (left + right) / 2, (bottom + top) / 2, top); 
    cilk_spawn update_quadtree_node(node->q3, collisionWorld, left, (left + right) / 2, bottom, (bottom + top) / 2); 
    update_quadtree_node(node->q4, collisionWorld, (left + right) / 2, right, bottom, (bottom + top) / 2); 
  } else {
    update_quadtree_node(node->q1, collisionWorld, (left + right) / 2, right, (bottom + top) / 2, top); 
    update_quadtree_node(node->q2, collisionWorld, left, (left + right) / 2, (bottom + top) / 2, top); 
    update_quadtree_node(node->q3, collisionWorld, left, (left + right) / 2, bottom, (bottom + top) / 2); 
    update_quadtree_node(node->q4, collisionWorld, (left + right) / 2, right, bottom, (bottom + top) / 2);  
  }

//  if (node->lineArray != NULL) {
//    free(node->lineArray);
//  }
//  node->lineArray = malloc(node->lineLinkedList.length * sizeof(Line *));
// Line* l1 = node->lineLinkedList.first;
//  for (size_t i = 0; i < node->lineLinkedList.length; i++) {
//    node->lineArray[i] = l1;
//    l1 = l1->next;
//  }
 
  cilk_sync;
}


void count_collision_from_tree_serial(QuadTree* node, CollisionWorld* collisionWorld) {
  if (node == NULL) {
    return;
  }
  count_collision_within_tree(node, collisionWorld);
 
   // Count the collisions with lines in the subtrees which the line belongs in
  count_collision_from_tree(node->q1, collisionWorld);
  count_collision_from_tree(node->q2, collisionWorld);
  count_collision_from_tree(node->q3, collisionWorld);
  count_collision_from_tree(node->q4, collisionWorld);

  Line * l1 = node->lineLinkedList.first;
  for (size_t i = 0; i < node->lineLinkedList.length; i++){
    count_collision_line_tree(l1, node, collisionWorld);
    l1 = l1->next;
  } 
}

void count_collision_from_tree(QuadTree* node, CollisionWorld* collisionWorld) {
//  const fasttime_t start_time = gettime();

  if (node == NULL) {
    return;
  }

  if (node->depth >= MAXDEPTH - 1) {
    count_collision_from_tree_serial(node, collisionWorld);
    return;
  }

// Count the collisions with lines in the subtrees which the line belongs in

//#ifdef CILKSCALE
//  cv_start();
//#endif
  cilk_spawn count_collision_from_tree(node->q1, collisionWorld);
  cilk_spawn count_collision_from_tree(node->q2, collisionWorld);
  cilk_spawn count_collision_from_tree(node->q3, collisionWorld);
  cilk_spawn count_collision_from_tree(node->q4, collisionWorld);

//#ifdef CILKSCALE
//  cv_stop();
//#endif
 
  cilk_spawn count_collision_within_tree(node, collisionWorld);

//  const fasttime_t mid_time = gettime();
//  printf("TIme Elapsed: %f, Midtime, Depth: %d \n",tdiff(start_time, mid_time), node->depth);

  cilk_spawn count_collision_line_list_tree(&node->lineLinkedList, node, collisionWorld);

//  printf("Depth %d Num Lines: %d\n", node->depth, node->lineLinkedList.length);
//  const fasttime_t end_time = gettime();
//  printf("TIme Elapsed: %f, Depth: %d \n",tdiff(start_time, end_time), node->depth);
}

void count_collision_line_list_tree(LineLinkedList * lineLinkedList, QuadTree * node, CollisionWorld * collisionWorld) {
  Line * l1 = lineLinkedList->first;

  for (size_t i = 0; i < lineLinkedList->length; i++){
    cilk_spawn count_collision_line_tree(l1, node, collisionWorld);
    l1 = l1->next;
  }
}
 
void count_collision_from_root(QuadTree* root, CollisionWorld* collisionWorld, IntersectionEventList * intersectionEventList) {
//  const fasttime_t start_time = gettime();
  CILK_C_REGISTER_REDUCER(intersectionEventListReducer);
  CILK_C_REGISTER_REDUCER(numLineLineCollisionReducer);   
 
  count_collision_from_tree(root, collisionWorld);

  cilk_sync;
 
  collisionWorld->numLineLineCollisions += numLineLineCollisionReducer.value;
  *intersectionEventList = intersectionEventListReducer.value;

  intersection_event_list_init(&REDUCER_VIEW(intersectionEventListReducer));
  num_line_line_collision_init(&REDUCER_VIEW(numLineLineCollisionReducer));

  CILK_C_UNREGISTER_REDUCER(numLineLineCollisionReducer); 
  CILK_C_UNREGISTER_REDUCER(intersectionEventListReducer);
  
//  const fasttime_t end_time = gettime();
//  printf("Total TIme Elapsed: %f",tdiff(start_time, end_time));
}

void count_collision_within_tree_parallel(QuadTree* node, CollisionWorld* collisionWorld) {
  size_t length = node->lineLinkedList.length;
//  printf("%d\n", length);

  Line ** lineList = malloc(length * sizeof(Line *));
  Line * curLine = node->lineLinkedList.first;
  for (int i=0; i < length; i++){
    lineList[i] = curLine;
    curLine = curLine->next;
  }

  cilk_for (int i = 0; i < length; i++) {
    Line * l1 = lineList[i];    
  
    // Count the collisions with lines in the current node  
    cilk_for (int j = 0; j < length; j++){
      if (j <= i) { continue;}
      Line * l2 = lineList[j];

      assert (l1->id != l2->id);    
      if ((l1->quad & l2->quad) != 0) {
        // Need to check the order of id
        if (compareLines(l1,l2) >= 0) {
          IntersectionType intersectionType =
              intersect(l2, l1, collisionWorld->timeStep);
          if (intersectionType != NO_INTERSECTION) {
            IntersectionEventList_appendNode(&REDUCER_VIEW(intersectionEventListReducer), l2, l1,
                                             intersectionType);
            REDUCER_VIEW(numLineLineCollisionReducer)++; 
          }
        } else {
          IntersectionType intersectionType =
              intersect(l1, l2, collisionWorld->timeStep);
          if (intersectionType != NO_INTERSECTION) {
            IntersectionEventList_appendNode(&REDUCER_VIEW(intersectionEventListReducer), l1, l2,
                                             intersectionType);
            REDUCER_VIEW(numLineLineCollisionReducer)++;
          }
        }
      }
    }
  }
  free(lineList);
}

void count_collision_within_tree(QuadTree* node, CollisionWorld* collisionWorld){
  assert(node != NULL);
  if (node->lineLinkedList.length > LOOPTHRESHOLD) {
    count_collision_within_tree_parallel(node, collisionWorld);
    return;
  }

  Line * l1 = node->lineLinkedList.first;

  for (int i = 0; i < node->lineLinkedList.length; i++){
    Line * l2 = l1->next;
  
    // Count the collisions with lines in the current node  
    for (int j = i+1; j < node->lineLinkedList.length; j++){
      assert (l1->id != l2->id);    

      if ((l1->quad & l2->quad) != 0) {
        // Need to check the order of id
        if (compareLines(l1,l2) >= 0) {
          IntersectionType intersectionType =
              intersect(l2, l1, collisionWorld->timeStep);
          if (intersectionType != NO_INTERSECTION) {
            IntersectionEventList_appendNode(&REDUCER_VIEW(intersectionEventListReducer), l2, l1,
                                             intersectionType);
            REDUCER_VIEW(numLineLineCollisionReducer)++; 
          }
        } else {
          IntersectionType intersectionType =
              intersect(l1, l2, collisionWorld->timeStep);
          if (intersectionType != NO_INTERSECTION) {
            IntersectionEventList_appendNode(&REDUCER_VIEW(intersectionEventListReducer), l1, l2,
                                             intersectionType);
            REDUCER_VIEW(numLineLineCollisionReducer)++;
          }
        }
      }
      l2 = l2->next;      
    }
    l1 = l1->next;
  }
}

void count_collision_line_tree(Line* l1, QuadTree* node, CollisionWorld* collisionWorld){
    if (l1->quad & 1) {
      count_collision_line_subtree(l1, node->q1, collisionWorld);
    }
    if (l1->quad >> 1 & 1) {
      count_collision_line_subtree(l1, node->q2, collisionWorld);
    }
    if (l1->quad >> 2 & 1) {
      count_collision_line_subtree(l1, node->q3, collisionWorld);
    }
    if (l1->quad >> 3 & 1){
      count_collision_line_subtree(l1, node->q4, collisionWorld);
    }
 
}

void count_collision_line_subtree(Line* l1, QuadTree* node, CollisionWorld* collisionWorld){
  if (node == NULL) {
    return;
  }

  count_collision_line_subtree(l1, node->q1, collisionWorld);
  count_collision_line_subtree(l1, node->q2, collisionWorld);
  count_collision_line_subtree(l1, node->q3, collisionWorld);
  count_collision_line_subtree(l1, node->q4, collisionWorld);

  Line * l2 = node->lineLinkedList.first;

  for (int i = 0; i < node->lineLinkedList.length; i++){
//    Line * l2 = node->lineArray[i];
    assert (l1->id != l2->id);    

    // Need to check the order of id
    if (compareLines(l1,l2) >= 0) {
      IntersectionType intersectionType =
          intersect(l2, l1, collisionWorld->timeStep);
      if (intersectionType != NO_INTERSECTION) {
        IntersectionEventList_appendNode(&REDUCER_VIEW(intersectionEventListReducer), l2, l1,
                                         intersectionType);
       REDUCER_VIEW(numLineLineCollisionReducer)++; 
        
      }
    } else {
      IntersectionType intersectionType =
          intersect(l1, l2, collisionWorld->timeStep);
      if (intersectionType != NO_INTERSECTION) {
        IntersectionEventList_appendNode(&REDUCER_VIEW(intersectionEventListReducer), l1, l2,
                                         intersectionType);
        REDUCER_VIEW(numLineLineCollisionReducer)++;
      }
    }
    l2 = l2->next;
  }
}
 
/*
// Counts the number of collisions between pairs of lines contained in this node
void count_collision_now(QuadTree* node, CollisionWorld* collisionWorld, IntersectionEventListReducer* intersection_event_list_reducer) {
  Line * l1 = node->lineLinkedList.first;

  for (int i = 0; i < node->lineLinkedList.length; i++){
    Line * l2 = l1->next;
  
    // Count the collisions with lines in the current node  
    for (int j = i+1; j < node->lineLinkedList.length; j++){
      assert (l1->id != l2->id);    

      // Need to check the order of id
      if (compareLines(l1,l2) >= 0) {
        IntersectionType intersectionType =
            intersect(l2, l1, collisionWorld->timeStep);
        if (intersectionType != NO_INTERSECTION) {
          IntersectionEventList_appendNode(&REDUCER_VIEW(*intersection_event_list_reducer), l2, l1,
                                           intersectionType);
          collisionWorld->numLineLineCollisions++;
        }
      } else {
        IntersectionType intersectionType =
            intersect(l1, l2, collisionWorld->timeStep);
        if (intersectionType != NO_INTERSECTION) {
          IntersectionEventList_appendNode(&REDUCER_VIEW(*intersection_event_list_reducer), l1, l2,
                                           intersectionType);
          collisionWorld->numLineLineCollisions++;
        }
      }
      l2 = l2->next;      
    }

    // Count the collisions with lines in the subtrees which the line belongs in
    if ((l1->quad & 1) && node->q1 != NULL) {
      count_collision_line_quadtree(l1, node->q1, collisionWorld, intersection_event_list_reducer);
    }
    if ((l1->quad >> 1 & 1) && node->q2 != NULL) {
      count_collision_line_quadtree(l1, node->q2, collisionWorld, intersection_event_list_reducer);
    }
    if ((l1->quad >> 2 & 1) && node->q3 != NULL) {
      count_collision_line_quadtree(l1, node->q3, collisionWorld, intersection_event_list_reducer);
    }
    if ((l1->quad >> 3 & 1) && node->q4 != NULL){
      count_collision_line_quadtree(l1, node->q4, collisionWorld, intersection_event_list_reducer);
    }
    
    l1 = l1->next;
  }
}

// counts number of collisions between l1, which is the line at the parent, and all lines at node, the subtree node
// updates IntersectionEventList with the collisions it detects
void count_collision_line_quadtree(Line *l1, QuadTree * node, CollisionWorld * collisionWorld, IntersectionEventListReducer* intersection_event_list_reducer) {
  Line * l2 = node->lineLinkedList.first;

  for (int j=0; j < node->lineLinkedList.length; j++) {
    assert (l1->id != l2->id);    
    
    // Need to check the order of id    
    if (compareLines(l1,l2) >= 0) {
      IntersectionType intersectionType =
          intersect(l2, l1, collisionWorld->timeStep);
      if (intersectionType != NO_INTERSECTION) {
        IntersectionEventList_appendNode(&REDUCER_VIEW(*intersection_event_list_reducer), l2, l1,
                                         intersectionType);
        collisionWorld->numLineLineCollisions++;
      }
    } else {
      IntersectionType intersectionType =
          intersect(l1, l2, collisionWorld->timeStep);
      if (intersectionType != NO_INTERSECTION) {
        IntersectionEventList_appendNode(&REDUCER_VIEW(*intersection_event_list_reducer), l1, l2,
                                         intersectionType);
        collisionWorld->numLineLineCollisions++;
      }
    }

    l2 = l2->next;
  }
}
*/
// Determines what quadrants a line has points in
void which_quadrant_line(Line * line, const vec_dimension left, const vec_dimension right, const vec_dimension bottom, const vec_dimension top, double time) { 
  // Note: bounding box may exceed this range when line-wall collision happens

  // Points at the parallelogram drawn by this line during given time
  int8_t s1, s2, s3, s4;
  
  vec_dimension x_middle = (left + right) / 2;
  vec_dimension y_middle = (bottom + top) / 2;
/*
  line->quad = 15;
  // When each statement is true, we rule out possibilities
  if (line->bounding_box.xmax < x_middle) {
    line->quad &= 6;
  } else if (line->bounding_box.xmin >= x_middle) {
    line->quad &= 9;
  }
  if (line->bounding_box.ymax < y_middle) {
    line->quad &= 3;
  } else if (line->bounding_box.ymin >= y_middle) {
    line->quad &= 12;
  }
*/
  s1 = which_quadrant_vec(&(line->p1), x_middle, y_middle);
  s2 = which_quadrant_vec(&(line->p2), x_middle, y_middle);
  s3 = which_quadrant_vec(&(line->bounding_box.target_p1), x_middle, y_middle);
  s4 = which_quadrant_vec(&(line->bounding_box.target_p2), x_middle, y_middle);

  int8_t bitarray[4] = {0, 0, 0, 0};
  bitarray[s1] = 1;
  bitarray[s2] = 1;
  bitarray[s3] = 1;
  bitarray[s4] = 1;

  line->quad = (bitarray[3] << 3) | (bitarray[2] << 2) | (bitarray[1] << 1) | bitarray[0];
}

/*
 Takes in a vector, vertical divider, and horiziontal divider to determine which
 which of 4 quadrants the vector is contained in:
 2 | 1
 ------
 3 | 4
*/
int which_quadrant_vec(const Vec * p, const vec_dimension x_middle, const vec_dimension y_middle) {
  int quadrant = (p->x < x_middle); // check the x-coordinate
  if (p->y < y_middle) { 
    quadrant ^= 3; // check the y-coordinate
  }
  assert(quadrant >= 0 && quadrant <= 3);
  return quadrant;
}

// Adds the line to the appropriate quadrants
inline void add_line_to_quadrant(Line* line, QuadTree* node) {
    int quad = line->quad;

    // Line belings in one sub-quadrant
    if (!(quad^1)) {
      add_line(&node->q1->lineLinkedList, line);
    } else if (!(quad^2)) { 
      add_line(&node->q2->lineLinkedList, line); 
    } else if (!(quad^4)) {
      add_line(&node->q3->lineLinkedList, line);
    } else if (!(quad^8)) {
      add_line(&node->q4->lineLinkedList, line);
    } else if (!(quad^3) || !(quad^6) || !(quad^12) || !(quad^9)) {
      // Line belongs to two adjacent sub-quadrants 
      add_line(&node->lineLinkedList, line);
    } else {
      // Opposite quadrants have points, so check all quadrants for collisions
      line->quad = 15;
      add_line(&node->lineLinkedList, line);
    }
}

// Recursively frees memory for this quadtree and initialized subtrees
void free_quadtree(QuadTree* root) {
  // root is never assumed to be null because we always mark the root as
  // initialized whenever we make the quadtree root or node 
  if (root->q1 != NULL) {
    free_quadtree(root->q1);
  }
  if (root->q2 != NULL) {
    free_quadtree(root->q2);
  }
  if (root->q3 != NULL) {
    free_quadtree(root->q3);
  }
  if (root->q4 != NULL) {
    free_quadtree(root->q4);
  }
  
//  if (root->lineArray != NULL){
//    free(root->lineArray);
//  }

  free(root);
}

