#ifndef QUADTREE_H_
#define QUADTREE_H_

#include <stdbool.h>

#include "./vec.h"
#include "./line.h"
#include "./intersection_event_list.h"
#include "./line_linked_list.h"

// Forward definition to prevent circulating
struct CollisionWorld;

typedef struct QuadTree QuadTree;

struct QuadTree{
  // Subtrees corresponding to the quadrants of this square
  // They go in the same order as the mathematical quadrants on a Cartesian graph:
  /*
          |   
      q2  |  q1
          |
     ------------
          |
      q3  |  q4
          |
  
  - lines < middle x line go to quadrants q2 or q3, >= go to quadrants q1 or q4
  - lines < middle y line go to quadrants q3 or q4, >= go to quadrants q2 or q1
  - lines whose parallelogram calculated from the velocity and timestep that are wholly contained within 
    one quadrant go to that one quadrant
  - lines whose parallelogram is is contained in two or more quadrants is stored in this QuadTree node
    (kept via updating "last" pointer instance variable). There are two cases:
    1. The line intersects with two quadrants which are adjacent. In this case, we compare this line with the
       lines also in the two quadrants' subtrees
    2. The line intersects with two quadrants which are in opposite corners. In this case, we comapre this line
       with lines in all four subtrees (because the parallelogram also will overlap a 3rd and possibly 4th 
       quadrant in addition to the two opposite quadrants
  */
  QuadTree * q1;
  QuadTree * q2;
  QuadTree * q3;
  QuadTree * q4;

  // This structure does not own Line * line, but keeps pointers to the first
  // and last lines. After construction, only the lines which cross into multiple
  // child quadrants are stored in the list starting at first and ending at last.
  // These lines are checked with each other for intersections, and also with the
  // lines in the appropriate quadtrees (determing these appropriate quadtrees is
  // detailed above)
  struct LineLinkedList lineLinkedList;

//  Line ** lineArray;
/*
  Line * first; // First element
  Line * last; // Last element to ease the addition of new line

  // The number of lines from first element to last element
  int line_length;
*/  
  // Depth of tree (how many pointers away from root)
  int depth;
  
  // Whether its subtree has been allocated memory
  bool initialized;
};

/*
// Unlink linked list of lines attached to this quadtree
void unlink_lines(QuadTree * root);

// Add one line to the linked list attached to this quadtree
void add_line(QuadTree * root, Line * line);
*/

// Makes a QuadTree root, calling make_quadtree_node for its 4 subchildren.
// Counts intersections with lines contained at the root and other lines at the root and lines
// contains in subtrees via call to count_collision_now
void make_quadtree_root(QuadTree* root, struct CollisionWorld* collisionWorld);

// Recursively builds a QuadTree node, and counts the number of collisions between lines contained at this node and other lines at this node
// and lines contained in subtrees
void make_quadtree_node(QuadTree* node, struct CollisionWorld* collisionWorld, const vec_dimension left, const vec_dimension right, const vec_dimension bottom, const vec_dimension top); 

// Recursively unlink quadtrees
void cleanup_quadtree(QuadTree * node);

// Updates a QuadTree root (with already allocated memory), calling update_quadtree_node for its 4 subchildren.
// It clears the lines contained at this root and recursively rebuilds the lines.
// Counts intersections with lines contained at the root and other lines at the root and lines
// contains in subtrees via call to count_collision_now
// Different from make_quadtree_root because this function does not reallocate memory for the root
void update_quadtree_root(QuadTree* root, struct CollisionWorld* collisionWorld);

// Recursively updates a QuadTree node, and counts the number of collisions between lines contained at this node and other lines at this node
// and lines contained in subtrees.
// Different from make_quadtree_node because memory for the node is already allocated beforehand.
// If memory is not allocated for subtrees, call make_quadtree_node.

void update_quadtree_node(QuadTree* node, struct CollisionWorld* collisionWorld, const vec_dimension left, const vec_dimension right, const vec_dimension bottom, const vec_dimension top);

void count_collision_from_root(QuadTree* root, struct CollisionWorld* collisoinWorld, IntersectionEventList* intersectionEventList);

void count_collision_from_tree(QuadTree* node, struct CollisionWorld* collisionWorld);

void count_collision_within_tree(QuadTree* node, struct CollisionWorld* collisionWorld);

void count_collision_line_list_tree(LineLinkedList * lineLinkedList, QuadTree * node, struct CollisionWorld* collisionWorld);

void count_collision_line_tree(Line * line, QuadTree * node, struct CollisionWorld* collisionWorld);

void count_collision_line_subtree(Line* line, QuadTree* node, struct CollisionWorld* collisionWorld);


// Count the number of collisions between (1) lines in the current node and 
// (2) lines in the current node and lines in it children
//
// Precondition: All line candidates that might collide with lines in the current node
// must be in itself or its direct children. This function does not check 
// descendents of child nodes.
//void count_collision_now(QuadTree* node, struct CollisionWorld* collisionWorld, IntersectionEventListReducer* intersection_event_list_reducer);

// Check whether line collides with lines linked to quadtree 
// and update the number of collisions tracked by collisionWorld
//void count_collision_line_quadtree(Line* line, QuadTree* quadtree, struct CollisionWorld* collisionWorld, IntersectionEventListReducer* intersection_event_list_reducer);

// Recursively free the Quadtree structure.
void free_quadtree(QuadTree * root);

// Set quad instance of line. Quad instance indicates location of vectors in line: 
// which quadrant these vectors belongs to in the specified region
//
// Quadrant:= 0: (+,+), 1: (-,+), 2: (-,-), 3: (+, -) in the coordinate system
void which_quadrant_line(Line * line, const vec_dimension left, const vec_dimension right, const vec_dimension bottom, const vec_dimension top, double time);

// Add this line to the quadrant(s) which it belongs, using line->quad calculated
// in which_quadrant_line. See QuadTree struct documentation for the 4 subtrees for
// definition of which quadrant(s) a line should be included in
void add_line_to_quadrant(Line * line, QuadTree * node);

// Given vector, return which quadrant this vector belongs to in the specified region
//
// Quadrant:= 0: (+,+), 1: (-,+), 2: (-,-), 3: (+, -) in the coordinate system
int which_quadrant_vec(const Vec * p, const vec_dimension x_middle, const vec_dimension y_middle);

#endif  // QUADTREE_H_
