#include "./line.h"
#include "./quadtree.h"
#include "./collision_world.h"
#include "./intersection_event_list.h"

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <cilk/cilk.h>

#define THRESHOLD 5
#define MAXDEPTH 5

// clears the Line points of root
void unlink_list(QuadTree * root){
  root->first = NULL;
  root->last = NULL;
  root->line_length = 0;
}

// adds a line to the root's Line pointers
void add_line(QuadTree * root, Line * line){
  line->next = NULL;
  root->line_length++;
  // Check whether ths line is the first line attached.
  if (root->first == NULL){
    root->first = line;
    root->last = line;
    return;
  }
  root->last->next = line;
  root->last = line;
}

// Makes a QuadTree root, calling make_quadtree_node for its 4 subchildren.
// Counts intersections with lines contained at the root and other lines at the root and lines
// contains in subtrees via call to count_collision_now
void make_quadtree_root(QuadTree* root, CollisionWorld* collisionWorld, IntersectionEventList* intersectionEventList) {  
  root->depth = 0;

  root->q1 = malloc(sizeof(QuadTree));
  root->q2 = malloc(sizeof(QuadTree));
  root->q3 = malloc(sizeof(QuadTree));
  root->q4 = malloc(sizeof(QuadTree));

  root->initialized = true;
  root->q1->initialized = false; 
  root->q2->initialized = false;
  root->q3->initialized = false;
  root->q4->initialized = false;
  
  root->q1->depth = 1;
  root->q2->depth = 1;
  root->q3->depth = 1;
  root->q4->depth = 1;

  // Initialize first and last in quadtress to be NULL
  unlink_list(root);
  unlink_list(root->q1);
  unlink_list(root->q2);
  unlink_list(root->q3);
  unlink_list(root->q4);

  for (int i = 0; i < collisionWorld->numOfLines; i++) {
    Line * line = collisionWorld->lines[i];
    which_quadrant_line(line, BOX_XMIN, BOX_XMAX, BOX_YMIN, BOX_YMAX, collisionWorld->timeStep); 
    add_line_to_quadrant(line, root);
  }

  // Count the possible number of collisions with the lines in the root
  count_collision_now(root, collisionWorld, intersectionEventList);
 
  // Recursively build quadtrees
  make_quadtree_node(root->q1, collisionWorld, intersectionEventList, (BOX_XMIN + BOX_XMAX) / 2, BOX_XMAX, (BOX_YMAX + BOX_YMIN) / 2, BOX_YMAX); 
  make_quadtree_node(root->q2, collisionWorld, intersectionEventList, BOX_XMIN, (BOX_XMIN + BOX_XMAX) / 2, (BOX_YMAX + BOX_YMIN) / 2, BOX_YMAX); 
  make_quadtree_node(root->q3, collisionWorld, intersectionEventList, BOX_XMIN, (BOX_XMIN + BOX_XMAX) / 2, BOX_YMIN, (BOX_YMAX + BOX_YMIN) / 2); 
  make_quadtree_node(root->q4, collisionWorld, intersectionEventList, (BOX_XMIN + BOX_XMAX) / 2, BOX_XMAX, BOX_YMIN, (BOX_YMAX + BOX_YMIN) / 2); 

}

// Recursively builds a QuadTree node, and counts the number of collisions between lines contained at this node and other lines at this node
// and lines contained in subtrees
void make_quadtree_node(QuadTree* node, CollisionWorld* collisionWorld, IntersectionEventList* intersectionEventList, const vec_dimension left, const vec_dimension right, const vec_dimension bottom, const vec_dimension top){  
  node->q1 = NULL;
  node->q2 = NULL;
  node->q3 = NULL;
  node->q4 = NULL;

  // If the number of lines is less then THRESHOLD
  // or we reach the maximum depth, stop build the tree and 
  // count the number of collisions
  if (node->line_length <= THRESHOLD || node->depth >= MAXDEPTH) {
    count_collision_now(node, collisionWorld, intersectionEventList);
    return;
  }

  // Initialization
  node->q1 = malloc(sizeof(QuadTree));
  node->q2 = malloc(sizeof(QuadTree));
  node->q3 = malloc(sizeof(QuadTree));
  node->q4 = malloc(sizeof(QuadTree));

  node->initialized = true;
  node->q1->initialized = false;
  node->q2->initialized = false;
  node->q3->initialized = false;
  node->q4->initialized = false;

  node->q1->depth = node->depth + 1;
  node->q2->depth = node->depth + 1;
  node->q3->depth = node->depth + 1;
  node->q4->depth = node->depth + 1;
 
  // Get the line attached first to the node before unlink it
  Line * line = node->first;
  int line_length = node->line_length;
  unlink_list(node);
  unlink_list(node->q1);
  unlink_list(node->q2);
  unlink_list(node->q3);
  unlink_list(node->q4);

  for(int i = 0; i < line_length; i++){
    Line * next = line->next;
    which_quadrant_line(line, left, right, bottom, top, collisionWorld->timeStep);

    add_line_to_quadrant(line, node);
    line = next;
  }

  // Count the possible number of collisions with the lines in the root
  count_collision_now(node, collisionWorld, intersectionEventList);

  // Recursively build quadtrees
  make_quadtree_node(node->q1, collisionWorld, intersectionEventList, (left + right) / 2, right, (bottom + top) / 2, top); 
  make_quadtree_node(node->q2, collisionWorld, intersectionEventList, left, (left + right) / 2, (bottom + top) / 2, top); 
  make_quadtree_node(node->q3, collisionWorld, intersectionEventList, left, (left + right) / 2, bottom, (bottom + top) / 2); 
  make_quadtree_node(node->q4, collisionWorld, intersectionEventList, (left + right) / 2, right, bottom, (bottom + top) / 2); 
}

// Updates a QuadTree root (with already allocated memory), calling update_quadtree_node for its 4 subchildren.
// It clears the lines contained at this root and recursively rebuilds the lines.
// Counts intersections with lines contained at the root and other lines at the root and lines
// contains in subtrees via call to count_collision_now
// Different from make_quadtree_root because this function does not reallocate memory for the root
void update_quadtree_root(QuadTree* root, CollisionWorld* collisionWorld, IntersectionEventList* intersectionEventList){
  unlink_list(root);
  unlink_list(root->q1);
  unlink_list(root->q2);
  unlink_list(root->q3);
  unlink_list(root->q4);

  for (int i = 0; i < collisionWorld->numOfLines; i++) {
    Line * line = collisionWorld->lines[i];
    which_quadrant_line(line, BOX_XMIN, BOX_XMAX, BOX_YMIN, BOX_YMAX, collisionWorld->timeStep);
 
    add_line_to_quadrant(line, root);
  }
  
  // Count the number of collisions with the lines in the root
  count_collision_now(root, collisionWorld, intersectionEventList);

  // Recursively update quadtrees
  update_quadtree_node(root->q1, collisionWorld, intersectionEventList, (BOX_XMIN + BOX_XMAX) / 2, BOX_XMAX, (BOX_YMAX + BOX_YMIN) / 2, BOX_YMAX); 
  update_quadtree_node(root->q2, collisionWorld, intersectionEventList, BOX_XMIN, (BOX_XMIN + BOX_XMAX) / 2, (BOX_YMAX + BOX_YMIN) / 2, BOX_YMAX); 
  update_quadtree_node(root->q3, collisionWorld, intersectionEventList, BOX_XMIN, (BOX_XMIN + BOX_XMAX) / 2, BOX_YMIN, (BOX_YMAX + BOX_YMIN) / 2); 
  update_quadtree_node(root->q4, collisionWorld, intersectionEventList, (BOX_XMIN + BOX_XMAX) / 2, BOX_XMAX, BOX_YMIN, (BOX_YMAX + BOX_YMIN) / 2); 
}


// Recursively updates a QuadTree node, and counts the number of collisions between lines contained at this node and other lines at this node
// and lines contained in subtrees.
// Different from make_quadtree_node because memory for the node is already allocated beforehand.
void update_quadtree_node(QuadTree* node, CollisionWorld* collisionWorld, IntersectionEventList* intersectionEventList, const vec_dimension left, const vec_dimension right, const vec_dimension bottom, const vec_dimension top){
  // If the node has not been initialized, call make_quadtree_node
  // to build tress from it.
  if (!node->initialized){
    make_quadtree_node(node, collisionWorld, intersectionEventList, left, right, bottom, top);
    return;
  }

  unlink_list(node->q1);
  unlink_list(node->q2);
  unlink_list(node->q3);
  unlink_list(node->q4);

  // If the number of lines is less then THRESHOLD
  // or we reach the maximum depth, stop build the tree and 
  // count the number of collisions
  if (node->line_length <= THRESHOLD || node->depth >= MAXDEPTH) {
    count_collision_now(node, collisionWorld, intersectionEventList);
    return;
  }

  // Get the first line attached before unlink it.
  Line * line = node->first;
  int line_length = node->line_length;
  unlink_list(node);

  for (int i = 0; i < line_length; i++) {
    which_quadrant_line(line, left, right, bottom, top, collisionWorld->timeStep);
    Line * next = line->next;

    add_line_to_quadrant(line, node);
    line = next;
  }

  // Count the number of collisions with the lines in this node
  count_collision_now(node, collisionWorld, intersectionEventList);

  // Recursively update quadtrees
  update_quadtree_node(node->q1, collisionWorld, intersectionEventList, (left + right) / 2, right, (bottom + top) / 2, top); 
  update_quadtree_node(node->q2, collisionWorld, intersectionEventList, left, (left + right) / 2, (bottom + top) / 2, top); 
  update_quadtree_node(node->q3, collisionWorld, intersectionEventList, left, (left + right) / 2, bottom, (bottom + top) / 2); 
  update_quadtree_node(node->q4, collisionWorld, intersectionEventList, (left + right) / 2, right, bottom, (bottom + top) / 2); 
}

// Counts the number of collisions between pairs of lines contained in this node
void count_collision_now(QuadTree* node, CollisionWorld* collisionWorld, IntersectionEventList* intersectionEventList){
  Line * l1 = node->first;

  for (int i = 0; i < node->line_length; i++){
    Line * l2 = l1->next;
  
    // Count the collisions with lines in the current node  
    for (int j = i+1; j < node->line_length; j++){
      assert (l1->id != l2->id);    

      // Need to check the order of id
      if (compareLines(l1,l2) >= 0) {
        IntersectionType intersectionType =
            intersect(l2, l1, collisionWorld->timeStep);
        if (intersectionType != NO_INTERSECTION) {
          IntersectionEventList_appendNode(intersectionEventList, l2, l1,
                                           intersectionType);
          collisionWorld->numLineLineCollisions++;
        }
      } else {
        IntersectionType intersectionType =
            intersect(l1, l2, collisionWorld->timeStep);
        if (intersectionType != NO_INTERSECTION) {
          IntersectionEventList_appendNode(intersectionEventList, l1, l2,
                                           intersectionType);
          collisionWorld->numLineLineCollisions++;
        }
      }
      l2 = l2->next;      
    }

    // Count the collisions with lines in the subtrees which the line belongs in
    if ((l1->quad & 1) && node->q1 != NULL) {
      count_collision_line_quadtree(l1, node->q1, collisionWorld, intersectionEventList);
    }
    if ((l1->quad >> 1 & 1) && node->q2 != NULL) {
      count_collision_line_quadtree(l1, node->q2, collisionWorld, intersectionEventList);
    }
    if ((l1->quad >> 2 & 1) && node->q3 != NULL) {
      count_collision_line_quadtree(l1, node->q3, collisionWorld, intersectionEventList);
    }
    if ((l1->quad >> 3 & 1) && node->q4 != NULL){
      count_collision_line_quadtree(l1, node->q4, collisionWorld, intersectionEventList);
    }
    
    l1 = l1->next;
  }
}

// counts number of collisions between l1, which is the line at the parent, and all lines at node, the subtree node
// updates IntersectionEventList with the collisions it detects
void count_collision_line_quadtree(Line *l1, QuadTree * node, CollisionWorld * collisionWorld, IntersectionEventList * intersectionEventList) {
  Line * l2 = node->first;

  for (int j=0; j < node->line_length; j++) {
    assert (l1->id != l2->id);    
    
    // Need to check the order of id    
    if (compareLines(l1,l2) >= 0) {
      IntersectionType intersectionType =
          intersect(l2, l1, collisionWorld->timeStep);
      if (intersectionType != NO_INTERSECTION) {
        IntersectionEventList_appendNode(intersectionEventList, l2, l1,
                                         intersectionType);
        collisionWorld->numLineLineCollisions++;
      }
    } else {
      IntersectionType intersectionType =
          intersect(l1, l2, collisionWorld->timeStep);
      if (intersectionType != NO_INTERSECTION) {
        IntersectionEventList_appendNode(intersectionEventList, l1, l2,
                                         intersectionType);
        collisionWorld->numLineLineCollisions++;
      }
    }

    l2 = l2->next;
  }
}

// Determines what quadrants a line has points in
void which_quadrant_line(Line * line, const vec_dimension left, const vec_dimension right, const vec_dimension bottom, const vec_dimension top, double time) {
  // Points should be within range
  assert(line->bounding_box.xmin >= left);
  assert(line->bounding_box.xmax <= right);
  assert(line->bounding_box.ymin >= bottom);
  assert(line->bounding_box.ymax <= top);


  // Points at the parallelogram drawn by this line during given time
  int8_t s1, s2, s3, s4;
  
  vec_dimension x_middle = (left + right) / 2;
  vec_dimension y_middle = (bottom + top) / 2;
/*
  line->quad = 15;
  // When each statement is true, we rule out possibilities
  if (line->bounding_box.xmax < x_middle) {
    line->quad &= 6;
  } else if (line->bounding_box.xmin >= x_middle) {
    line->quad &= 9;
  }
  if (line->bounding_box.ymax < y_middle) {
    line->quad &= 3;
  } else if (line->bounding_box.ymin >= y_middle) {
    line->quad &= 12;
  }
*/
  s1 = which_quadrant_vec(&(line->p1), x_middle, y_middle);
  s2 = which_quadrant_vec(&(line->p2), x_middle, y_middle);
  s3 = which_quadrant_vec(&(line->bounding_box.target_p1), x_middle, y_middle);
  s4 = which_quadrant_vec(&(line->bounding_box.target_p2), x_middle, y_middle);

  int8_t bitarray[4] = {0, 0, 0, 0};
  bitarray[s1] = 1;
  bitarray[s2] = 1;
  bitarray[s3] = 1;
  bitarray[s4] = 1;

  line->quad = (bitarray[3] << 3) | (bitarray[2] << 2) | (bitarray[1] << 1) | bitarray[0];
}

/*
 Takes in a vector, vertical divider, and horiziontal divider to determine which
 which of 4 quadrants the vector is contained in:
 2 | 1
 ------
 3 | 4
*/
int which_quadrant_vec(const Vec * p, const vec_dimension x_middle, const vec_dimension y_middle) {
  int quadrant = (p->x < x_middle); // check the x-coordinate
  if (p->y < y_middle) { 
    quadrant ^= 3; // check the y-coordinate
  }
  assert(quadrant >= 0 && quadrant <= 3);
  return quadrant;
}

// Adds the line to the appropriate quadrants
inline void add_line_to_quadrant(Line* line, QuadTree* node) {
    int quad = line->quad;
    // Line belings in one sub-quadrant
    if (!(quad^1)) {
      add_line(node->q1, line);
    } else if (!(quad^2)) { 
      add_line(node->q2, line); 
    } else if (!(quad^4)) {
      add_line(node->q3, line);
    } else if (!(quad^8)) {
      add_line(node->q4, line);
    } else if (!(quad^3) || !(quad^6) || !(quad^12) || !(quad^9)) {
      // Line belongs to two adjacent sub-quadrants 
      add_line(node, line);
    } else {
      // Opposite quadrants have points, so check all quadrants for collisions
      line->quad = 15;
      add_line(node, line);
    }
}

// Recursively frees memory for this quadtree and initialized subtrees
void free_quadtree(QuadTree* root) {
  if (root->q1 != NULL) {
    free_quadtree(root->q1);
  }
  if (root->q2 != NULL) {
    free_quadtree(root->q2);
  }
  if (root->q3 != NULL) {
    free_quadtree(root->q3);
  }
  if (root->q4 != NULL) {
    free_quadtree(root->q4);
  }
  free(root);
}

